package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerFilterOnlyOut
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.matter.baselineComplexityReplicateTicks
import ru.dbotthepony.mc.otm.matter.getMatterValue
import ru.dbotthepony.mc.otm.menu.MatterReplicatorMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.*

class MatterReplicatorBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryWorkerBlockEntity<MatterReplicatorBlockEntity.ReplicatorJob>(MBlockEntities.MATTER_REPLICATOR, p_155229_, p_155230_, {
		try {
			ReplicatorJob(it)
		} catch(err: NoSuchElementException) {
			null
		}
	}), IMatterGraphNode, IDroppableContainer {

	class ReplicatorJob : ItemJob {
		val matterPerTick: ImpreciseFraction
		val task: ReplicationTask
		var matterValue: ImpreciseFraction
		val pattern: PatternState?
		val asDust: Boolean

		constructor(tag: CompoundTag) : super(tag) {
			matterPerTick = tag.getImpreciseFraction(MATTER_PER_TICK_KEY)
			matterValue = tag.getImpreciseFraction(MATTER_VALUE_KEY)
			pattern = tag.map(PATTERN_KEY, PatternState::deserializeNBT)
			asDust = tag.getBoolean(AS_DUST_KEY)
			task = tag.map(TASK_KEY, ReplicationTask::deserializeNBT) ?: throw NoSuchElementException("Unable to deserialize matter task")
		}

		constructor(
			itemStack: ItemStack,
			matterPerTick: ImpreciseFraction,
			task: ReplicationTask,
			matterValue: ImpreciseFraction,
			pattern: PatternState?,
			asDust: Boolean,
			ticks: Double,
		) : super(itemStack, ticks, BASE_CONSUMPTION) {
			this.matterPerTick = matterPerTick
			this.task = task
			this.matterValue = matterValue
			this.pattern = pattern
			this.asDust = asDust
		}

		override fun serializeNBT(): CompoundTag {
			return super.serializeNBT().also {
				it[MATTER_PER_TICK_KEY] = this.matterPerTick
				it[TASK_KEY] = this.task.serializeNBT()
				it[MATTER_VALUE_KEY] = this.matterValue

				if (this.pattern != null)
					it[PATTERN_KEY] = this.pattern.serializeNBT()

				it[AS_DUST_KEY] = this.asDust
			}
		}

		companion object {
			const val MATTER_PER_TICK_KEY = "matterPerTick"
			const val MATTER_VALUE_KEY = "matterValue"
			const val PATTERN_KEY = "pattern"
			const val AS_DUST_KEY = "asDust"
			const val TASK_KEY = "task"
		}
	}

	override val energy = WorkerEnergyStorage(this::powerLevelUpdated, STORAGE, MAX_IO)
	override val matterNode = Graph6Node<IMatterGraphNode>(this)
	private val resolverNode = LazyOptional.of { this }

	val matter = MatterHandlerImpl(
		this::matterLevelUpdated,
		MatterDirection.RECEIVE,
		ImpreciseFraction(2)
	)

	val container = MatteryContainer(this::itemContainerUpdated, 5)
	private val itemHandler = container.handler(MatteryContainerFilterOnlyOut)

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override val droppableContainer: Container
		get() = container

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterReplicatorMenu(containerID, inventory, this)
	}

	override fun onJobFinish(job: ReplicatorJob): Status {
		if (job.asDust) {
			job.matterValue = moveMatterAsDustIntoContainer(job.matterValue, container, OUTPUT_DUST_MAIN, OUTPUT_DUST_STACKING)

			if (!job.matterValue.isZero) {
				return Status.FAILURE_WAIT
			}

			(matterNode.graph as MatterNetworkGraph?)?.notifyTaskCompletion(job.task.id)
			return Status.SUCCESS
		}

		if (!container.fullyAddItem(job.itemStack, FIRST_ACTUAL_OUTPUT_SLOT .. LAST_ACTUAL_OUTPUT_SLOT)) {
			return Status.FAILURE_ITEM
		}

		(matterNode.graph as MatterNetworkGraph?)?.notifyTaskCompletion(job.task.id)
		return Status.SUCCESS
	}

	override fun onMatterTaskCreated(task: IReplicationTask<*>) {
		if (idleReason == IdleReason.OBSERVING) {
			isIdling = false
		}
	}

	override fun <T : IReplicationTask<*>> onMatterTaskUpdated(newState: T, oldState: T) {
		if (idleReason == IdleReason.OBSERVING) {
			isIdling = false
		}
	}

	override fun onPatternAdded(state: IPatternState) {
		if (idleReason == IdleReason.OBSERVING) {
			isIdling = false
		}
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, matterNode)
	}

	override fun computeNextJob(): Pair<ReplicatorJob?, IdleReason?> {
		if (energy.batteryLevel < BASE_CONSUMPTION) {
			return null to IdleReason.POWER
		}

		val graph = matterNode.graph as MatterNetworkGraph? ?: return null to null
		val allocation = graph.allocateTask(simulate = false) ?: return null to IdleReason.OBSERVING
		val stack = allocation.task.stack(1)
		val matter = getMatterValue(stack)

		// ????????
		if (matter.isZero) return null to null

		val ticks = matter.complexity * baselineComplexityReplicateTicks

		return ReplicatorJob(
			itemStack = stack,
			matterPerTick = matter.value / ticks,
			task = allocation.task.asImmutable(),
			matterValue = matter.value,
			pattern = allocation.pattern?.asImmutable(),
			asDust = (level?.random?.nextDouble() ?: 1.0) > (allocation.pattern?.researchPercent ?: 2.0),
			ticks = ticks,
		) to null
	}

	override fun onWorkTick(requiredPower: ImpreciseFraction, extractedPower: ImpreciseFraction, ticksAdvanced: Double): Status {
		val drainPerTick = currentJob!!.matterPerTick * ticksAdvanced
		val graph = matterNode.graph as MatterNetworkGraph? ?: return Status.FAILURE_WAIT_FAST

		if (matter.extractMatterInner(drainPerTick, true) < drainPerTick) {
			// в машине недостаточно материи

			if (drainPerTick > matter.maxStoredMatter) {
				// в тик требуется больше материи, чем её может хранить репликатор
				val toExtract = drainPerTick - matter.extractMatterInner(drainPerTick, true)
				val drain = graph.extractMatter(toExtract, true)

				if (drain != toExtract) {
					// недостаточно материи в сети
					return Status.FAILURE_MATTER
				}

				// достаточно материи в сети + внутри машины
				matter.extractMatterInner(drainPerTick, false)
				graph.extractMatter(drain, false)
				return Status.SUCCESS
			} else {
				// в тик требуется меньше материи, чем её может хранить репликатор
				// примем из сети недостающее количество бака материи, или 200 тиков репликации, что меньше
				val drain = graph.extractMatter(matter.missingMatter.coerceAtMost(drainPerTick * DRAIN_MULT), true)

				if (drain.isZero) {
					// в сети нет материи
					return Status.FAILURE_MATTER
				}

				val received = matter.receiveMatterOuter(drain, false)
				graph.extractMatter(received, false)

				// получили материю, проверяем возможность работы
				if (matter.extractMatterInner(drainPerTick, true) >= drainPerTick) {
					matter.extractMatterInner(drainPerTick, false)
					return Status.SUCCESS
				} else {
					// :(
					return Status.FAILURE_WAIT
				}
			}
		}

		// в машине достаточно материи
		matter.extractMatterInner(drainPerTick, false)
		return Status.SUCCESS
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[INVENTORY_KEY] = container.serializeNBT()
		nbt[MATTER_STORAGE_KEY] = matter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.map(INVENTORY_KEY, container::deserializeNBT)
		nbt.map(MATTER_STORAGE_KEY, matter::deserializeNBT)
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		itemHandler.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		itemHandler.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap == MatteryCapability.MATTER_NODE) return resolverNode.cast()
			if (cap == ForgeCapabilities.ITEM_HANDLER) return itemHandler.get().cast()
		}

		return super.getCapability(cap, side)
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_replicator")
		private val BASE_CONSUMPTION = ImpreciseFraction(400)
		private val DRAIN_MULT = ImpreciseFraction(200)
		private val STORAGE = ImpreciseFraction(200_000)
		private val MAX_IO = ImpreciseFraction(4_000)

		const val FIRST_ACTUAL_OUTPUT_SLOT = 0
		const val LAST_ACTUAL_OUTPUT_SLOT = 2
		const val OUTPUT_DUST_MAIN = 3
		const val OUTPUT_DUST_STACKING = 4
	}
}
