
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.capability

import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.IEnergyStorage
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.compat.mekanism.Mattery2MekanismEnergyWrapper
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.formatPower
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.tagNotNull

sealed interface IEnergyStorageImpl {
	val maxInput: ImpreciseFraction?
	val maxOutput: ImpreciseFraction?
	val direction: FlowDirection
}

private fun batteryLevel(it: IEnergyStorage, tooltips: MutableList<Component>) {
	tooltips.add(TranslatableComponent(
		"otm.item.power.storage",
		it.energyStored.formatPower(),
		it.maxEnergyStored.formatPower()
	).withStyle(ChatFormatting.GRAY))
}

private fun batteryLevel(it: IMatteryEnergyStorage, tooltips: MutableList<Component>) {
	tooltips.add(TranslatableComponent(
		"otm.item.power.storage",
		it.batteryLevel.formatPower(),
		it.maxBatteryLevel.formatPower()
	).withStyle(ChatFormatting.GRAY))

	if (it is IEnergyStorageImpl) {
		when (it.direction) {
			FlowDirection.INPUT -> {
				if (it.maxInput != null) {
					tooltips.add(TranslatableComponent("otm.item.power.throughput_mono", it.maxInput!!.formatPower()).withStyle(ChatFormatting.GRAY))
				} else {
					tooltips.add(TranslatableComponent("otm.item.power.throughput_mono", TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY)).withStyle(ChatFormatting.GRAY))
				}
			}

			FlowDirection.OUTPUT -> {
				if (it.maxOutput != null) {
					tooltips.add(TranslatableComponent("otm.item.power.throughput_mono", it.maxOutput!!.formatPower()).withStyle(ChatFormatting.GRAY))
				} else {
					tooltips.add(TranslatableComponent("otm.item.power.throughput_mono", TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY)).withStyle(ChatFormatting.GRAY))
				}
			}

			FlowDirection.BI_DIRECTIONAL -> {
				val maxInput = it.maxInput
				val maxOutput = it.maxOutput

				if (maxInput != null && maxOutput != null) {
					tooltips.add(TranslatableComponent("otm.item.power.throughput", maxInput.formatPower(), maxOutput.formatPower()).withStyle(ChatFormatting.GRAY))
				} else if (maxInput != null) {
					tooltips.add(TranslatableComponent("otm.item.power.throughput",
						maxInput.formatPower(),
						TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY)
					).withStyle(ChatFormatting.GRAY))
				} else if (maxOutput != null) {
					tooltips.add(TranslatableComponent("otm.item.power.throughput",
						TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY),
						maxOutput.formatPower(),
					).withStyle(ChatFormatting.GRAY))
				} else {
					tooltips.add(TranslatableComponent("otm.item.power.throughput",
						TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY),
						TranslatableComponent("item.power.infinity").withStyle(ChatFormatting.GRAY),
					).withStyle(ChatFormatting.GRAY))
				}
			}
		}
	}
}

sealed class ItemEnergyStorageImpl(
	final override val direction: FlowDirection,
	protected val itemStack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction?,
	maxOutput: ImpreciseFraction?,
	val initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : IMatteryEnergyStorage, ICapabilityProvider, IEnergyStorageImpl {
	final override var maxInput: ImpreciseFraction? = maxInput
		protected set

	final override var maxOutput: ImpreciseFraction? = maxOutput
		protected set

	private val resolver = LazyOptional.of { this }
	private val resolverMekanism = if (isMekanismLoaded) LazyOptional.of { Mattery2MekanismEnergyWrapper(this) } else null

	override fun <T : Any> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (cap == ForgeCapabilities.ENERGY || cap == MatteryCapability.ENERGY) {
			return resolver.cast()
		} else if (cap == MatteryCapability.MEKANISM_ENERGY) {
			return resolverMekanism?.cast() ?: LazyOptional.empty()
		}

		return LazyOptional.empty()
	}

	override var maxBatteryLevel: ImpreciseFraction = maxBatteryLevel
		protected set

	override var batteryLevel: ImpreciseFraction
		get() = itemStack.tag?.map(ENERGY_KEY, ImpreciseFraction::deserializeNBT) ?: initialBatteryLevel
		protected set(value) {
			itemStack.tagNotNull[ENERGY_KEY] = value.serializeNBT()
		}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == FlowDirection.INPUT)
			return ImpreciseFraction.ZERO

		return extractEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == FlowDirection.OUTPUT)
			return ImpreciseFraction.ZERO

		return receiveEnergyInner(howMuch, simulate)
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive || itemStack.count != 1)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxOutput = maxOutput

		if (maxOutput != null) {
			howMuch = howMuch.coerceAtMost(maxOutput)
		}

		val batteryLevel = batteryLevel

		if (!batteryLevel.isPositive)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel - howMuch).moreThanZero()
		val diff = (batteryLevel - newLevel)

		if (!simulate && batteryLevel != newLevel) {
			this.batteryLevel = newLevel
		}

		return diff
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive || itemStack.count != 1)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxInput = maxInput

		if (maxInput != null) {
			howMuch = howMuch.coerceAtMost(maxInput)
		}

		val batteryLevel = batteryLevel

		if (batteryLevel >= maxBatteryLevel)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel + howMuch).coerceAtMost(maxBatteryLevel)
		val diff = (newLevel - batteryLevel)

		if (!simulate && batteryLevel != newLevel) {
			this.batteryLevel = newLevel
		}

		return diff
	}

	override fun canExtract(): Boolean {
		return direction != FlowDirection.INPUT
	}

	override fun canReceive(): Boolean {
		return direction != FlowDirection.OUTPUT
	}

	companion object {
		const val ENERGY_KEY = "energy"

		fun appendHoverText(itemStack: ItemStack, level: Level?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			appendHoverText(itemStack, tooltips)
		}

		fun appendHoverText(itemStack: ItemStack, tooltips: MutableList<Component>) {
			val energy = itemStack.energy ?: return

			if (energy is IMatteryEnergyStorage) {
				batteryLevel(energy, tooltips)
			} else {
				batteryLevel(energy, tooltips)
			}
		}
	}
}

open class EnergyConsumerItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(FlowDirection.INPUT, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

open class EnergyProducerItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(FlowDirection.OUTPUT, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

open class EnergyCapacitorItem(
	stack: ItemStack,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction? = null,
	maxOutput: ImpreciseFraction? = maxInput,
	initialBatteryLevel: ImpreciseFraction = ImpreciseFraction.ZERO
) : ItemEnergyStorageImpl(FlowDirection.BI_DIRECTIONAL, stack, maxBatteryLevel, maxInput, maxOutput, initialBatteryLevel)

sealed class BlockEnergyStorageImpl constructor(
	protected val listener: () -> Unit,
	final override val direction: FlowDirection,
	maxBatteryLevel: ImpreciseFraction,
	maxInput: ImpreciseFraction?,
	maxOutput: ImpreciseFraction?,
) : IMatteryEnergyStorage, INBTSerializable<CompoundTag>, IEnergyStorageImpl {
	final override var maxInput: ImpreciseFraction? = maxInput
		protected set
	final override var maxOutput: ImpreciseFraction? = maxOutput
		protected set

	override var maxBatteryLevel: ImpreciseFraction = maxBatteryLevel
		protected set(value) {
			if (value != field) {
				field = value
				listener.invoke()
			}
		}

	override var batteryLevel = ImpreciseFraction.ZERO
		protected set(value) {
			if (value != field) {
				field = value
				listener.invoke()
			}
		}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == FlowDirection.INPUT)
			return ImpreciseFraction.ZERO

		return extractEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (direction == FlowDirection.OUTPUT)
			return ImpreciseFraction.ZERO

		return receiveEnergyInner(howMuch, simulate)
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxOutput = maxOutput

		if (maxOutput != null) {
			howMuch = howMuch.coerceAtMost(maxOutput)
		}

		if (!batteryLevel.isPositive)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel - howMuch).moreThanZero()
		val diff = (batteryLevel - newLevel)

		if (!simulate && batteryLevel != newLevel) {
			batteryLevel = newLevel
		}

		return diff
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (!howMuch.isPositive)
			return ImpreciseFraction.ZERO

		@Suppress("NAME_SHADOWING")
		var howMuch = howMuch
		val maxInput = maxInput

		if (maxInput != null) {
			howMuch = howMuch.coerceAtMost(maxInput)
		}

		if (batteryLevel >= maxBatteryLevel)
			return ImpreciseFraction.ZERO

		val newLevel = (batteryLevel + howMuch).coerceAtMost(maxBatteryLevel)
		val diff = (newLevel - batteryLevel)

		if (!simulate && batteryLevel != newLevel) {
			batteryLevel = newLevel
		}

		return diff
	}

	override fun canExtract(): Boolean {
		return direction != FlowDirection.INPUT
	}

	override fun canReceive(): Boolean {
		return direction != FlowDirection.OUTPUT
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()
		tag[ENERGY_STORED_KEY] = batteryLevel.serializeNBT()
		tag[ENERGY_STORED_MAX_KEY] = maxBatteryLevel.serializeNBT()

		maxInput?.let { tag[MAX_INPUT_KEY] = it.serializeNBT() }
		maxOutput?.let { tag[MAX_OUTPUT_KEY] = it.serializeNBT() }

		return tag
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		batteryLevel = nbt.map(ENERGY_STORED_KEY, ImpreciseFraction.Companion::deserializeNBT) ?: ImpreciseFraction.ZERO
		maxBatteryLevel = nbt.map(ENERGY_STORED_MAX_KEY, ImpreciseFraction.Companion::deserializeNBT) ?: ImpreciseFraction.ZERO
		maxInput = nbt.map(MAX_INPUT_KEY, ImpreciseFraction.Companion::deserializeNBT)
		maxOutput = nbt.map(MAX_OUTPUT_KEY, ImpreciseFraction.Companion::deserializeNBT)
	}

	var resolver: LazyOptional<IMatteryEnergyStorage> = LazyOptional.of { this }
		private set

	fun invalidate() {
		resolver.invalidate()
	}

	fun revive() {
		resolver = LazyOptional.of { this }
	}

	companion object {
		val DEFAULT_MAX_IO = ImpreciseFraction(200)
		val DEFAULT_MAX_CAPACITY = ImpreciseFraction(60_000)

		const val ENERGY_STORED_KEY = "energy_stored"
		const val ENERGY_STORED_MAX_KEY = "energy_stored_max"
		const val MAX_INPUT_KEY = "max_input"
		const val MAX_OUTPUT_KEY = "max_output"
	}
}

open class WorkerEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, FlowDirection.INPUT, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)

	companion object {
		fun appendHoverText(itemStack: ItemStack, tooltips: MutableList<Component>) {
			val tag = (itemStack.tag?.get(BlockItem.BLOCK_ENTITY_TAG) as? CompoundTag)?.get(MatteryBlockEntity.ENERGY_KEY) as? CompoundTag ?: return
			val cap = WorkerEnergyStorage({})
			cap.deserializeNBT(tag)
			batteryLevel(cap, tooltips)
		}

		fun appendHoverText(itemStack: ItemStack, blockGetter: BlockGetter?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			return appendHoverText(itemStack, tooltips)
		}
	}
}

open class GeneratorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, FlowDirection.OUTPUT, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)

	companion object {
		fun appendHoverText(itemStack: ItemStack, tooltips: MutableList<Component>) {
			val tag = (itemStack.tag?.get(BlockItem.BLOCK_ENTITY_TAG) as? CompoundTag)?.get(MatteryBlockEntity.ENERGY_KEY) as? CompoundTag ?: return
			val cap = GeneratorEnergyStorage({})
			cap.deserializeNBT(tag)
			batteryLevel(cap, tooltips)
		}

		fun appendHoverText(itemStack: ItemStack, blockGetter: BlockGetter?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			return appendHoverText(itemStack, tooltips)
		}
	}
}

open class CapacitorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction? = maxInput
) : BlockEnergyStorageImpl(listener, FlowDirection.BI_DIRECTIONAL, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction? = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction? = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)

	companion object {
		fun appendHoverText(itemStack: ItemStack, tooltips: MutableList<Component>) {
			val tag = (itemStack.tag?.get(BlockItem.BLOCK_ENTITY_TAG) as? CompoundTag)?.get(MatteryBlockEntity.ENERGY_KEY) as? CompoundTag ?: return
			val cap = CapacitorEnergyStorage({})
			cap.deserializeNBT(tag)
			batteryLevel(cap, tooltips)
		}

		fun appendHoverText(itemStack: ItemStack, blockGetter: BlockGetter?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			return appendHoverText(itemStack, tooltips)
		}
	}
}
