package ru.dbotthepony.mc.otm.container

import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.IItemHandler

interface MatteryContainerFilter {
	fun canInsert(slot: Int, stack: ItemStack): Boolean
	fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean
}

object MatteryContainerFilterOnlyIn : MatteryContainerFilter {
	override fun canInsert(slot: Int, stack: ItemStack): Boolean {
		return true
	}

	override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
		return false
	}
}

object MatteryContainerFilterOnlyOut : MatteryContainerFilter {
	override fun canInsert(slot: Int, stack: ItemStack): Boolean {
		return false
	}

	override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
		return true
	}
}

object MatteryContainerFilterBoth : MatteryContainerFilter {
	override fun canInsert(slot: Int, stack: ItemStack): Boolean {
		return true
	}

	override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
		return true
	}
}

class MatteryContainerHandler @JvmOverloads internal constructor(
	private val container: MatteryContainer,
	private val filter: MatteryContainerFilter = MatteryContainerFilterBoth,
) : IItemHandler {
	private var handler = LazyOptional.of<IItemHandler> { this }

	fun get(): LazyOptional<IItemHandler> {
		return handler
	}

	fun invalidate() {
		handler.invalidate()
	}

	fun revive() {
		handler = LazyOptional.of { this }
	}

	override fun getSlots() = container.containerSize
	override fun getStackInSlot(slot: Int) = container[slot]

	override fun insertItem(slot: Int, stack: ItemStack, simulate: Boolean): ItemStack {
		if (!filter.canInsert(slot, stack))
			return stack

		val localStack = container[slot]

		if (localStack.isEmpty) {
			if (!simulate) {
				val copy = stack.copy()
				container.setChanged(slot, copy, ItemStack.EMPTY)
				container.setItem(slot, copy)
			}

			return ItemStack.EMPTY
		} else if (localStack.isStackable && localStack.maxStackSize > localStack.count && ItemStack.isSameItemSameTags(localStack, stack)) {
			val newCount = Math.min(localStack.maxStackSize, localStack.count + stack.count)
			val diff = newCount - localStack.count

			if (diff != 0) {
				if (!simulate) {
					val copy = localStack.copy()
					localStack.grow(diff)
					container.setChanged(slot, localStack, copy)
				}

				val copy = stack.copy()
				copy.shrink(diff)
				return copy
			}
		}

		return stack
	}

	override fun extractItem(slot: Int, amount: Int, simulate: Boolean): ItemStack {
		if (amount == 0)
			return ItemStack.EMPTY

		require(amount >= 0) { "Can not extract negative amount of items" }

		val localStack = container.getItem(slot)
		if (localStack.isEmpty) return ItemStack.EMPTY
		if (!filter.canExtract(slot, amount, localStack)) return ItemStack.EMPTY

		val minimal = Math.min(amount, localStack.count)
		val copy = localStack.copy()
		copy.count = minimal

		if (!simulate) {
			val copy = localStack.copy()
			localStack.shrink(minimal)
			container.setChanged(slot, localStack, copy)
		}

		return copy
	}

	override fun getSlotLimit(slot: Int): Int {
		return container.maxStackSize
	}

	override fun isItemValid(slot: Int, stack: ItemStack): Boolean {
		return filter.canInsert(slot, stack)
	}
}
