package ru.dbotthepony.mc.otm.core

import com.mojang.math.Matrix3f
import com.mojang.math.Matrix4f
import com.mojang.math.Vector3f
import com.mojang.math.Vector4f
import org.jetbrains.annotations.Contract
import java.nio.FloatBuffer

fun Matrix3f.asMutableMatrix(): MutableMatrix = MutableMatrix(this)
fun Matrix4f.asMutableMatrix(): MutableMatrix = MutableMatrix(this)

fun Matrix4f.rotate(angle: IAngle): Matrix4f {
	multiply(angle.rotationXYZW().toMatrix4f())
	return this
}

fun Matrix4f.translate(vector: Vector) {
	translate(-vector.asVector3f())
}

fun Matrix4f.rotateAroundPoint(point: Vector, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val p = point.asVector3f()
	translate(-p)
	multiply(axis.rotateAroundThis(rotation, isDegrees))
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector, axis: Vector3f, rotation: Float, isDegrees: Boolean = false) {
	val p = point.asVector3f()
	translate(-p)
	multiply(if (isDegrees) axis.rotationDegrees(rotation) else axis.rotation(rotation))
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector, rotation: IAngle) {
	val p = point.asVector3f()
	translate(-p)
	multiply(rotation.rotationXYZW().toMatrix4f())
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	translate(-point)
	multiply(axis.rotateAroundThis(rotation, isDegrees))
	translate(-point)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, rotation: IAngle) {
	translate(-point)
	multiply(rotation.rotationXYZW().toMatrix4f())
	translate(-point)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, axis: Vector3f, rotation: Float, isDegrees: Boolean = false) {
	translate(-point)
	multiply(if (isDegrees) axis.rotationDegrees(rotation) else axis.rotation(rotation))
	translate(-point)
}

private val floatBuff3 = FloatBuffer.allocate(3 * 3)
private val floatBuff4 = FloatBuffer.allocate(4 * 4)

fun Matrix4f.getTranslation(): Vector3f {
	store(floatBuff4)

	return Vector3f(floatBuff4[0 + 3 * 4], floatBuff4[1 + 3 * 4], floatBuff4[2 + 3 * 4])
}

fun Matrix4f.getTranslation4(): Vector4f {
	store(floatBuff4)

	return Vector4f(floatBuff4[0 + 3 * 4], floatBuff4[1 + 3 * 4], floatBuff4[2 + 3 * 4], floatBuff4[3 + 3 * 4])
}

class MutableMatrix(val rows: Int, val columns: Int = 1) {
	private val buffer = Array(columns * rows) {0.0}

	constructor(matrix3f: Matrix3f) : this(3, 3) {
		matrix3f.store(floatBuff3)
		load(floatBuff3)
	}

	constructor(matrix4f: Matrix4f) : this(4, 4) {
		matrix4f.store(floatBuff4)
		load(floatBuff4)
	}

	constructor(other: MutableMatrix) : this(other.rows, other.columns) {
		for (i in 0 .. buffer.size) {
			buffer[i] = other.buffer[i]
		}
	}

	fun identity(): MutableMatrix {
		check(columns == rows) { "Matrix is not cubic ($rows by $columns)" }

		for (column in 0 until columns) {
			for (row in 0 until rows) {
				buffer[row + column * rows] = if (column == row) 1.0 else 0.0
			}
		}

		return this
	}

	fun identityFast(): MutableMatrix {
		check(columns == rows) { "Matrix is not cubic ($rows by $columns)" }

		for (i in 0 until rows) {
			buffer[i + i * columns] = 1.0
		}

		return this
	}

	fun toMatrix3f(): Matrix3f {
		check(rows == columns && rows == 3) { "This matrix has wrong dimensions: $rows by $columns" }
		store(floatBuff3)
		return Matrix3f().also { it.load(floatBuff3) }
	}

	fun toMatrix4f(): Matrix4f {
		check(rows == columns && rows == 4) { "This matrix has wrong dimensions: $rows by $columns" }
		store(floatBuff4)
		return Matrix4f().also { it.load(floatBuff4) }
	}

	fun store(buff: FloatBuffer) {
		for (row in 0 until rows) {
			for (column in 0 until columns) {
				buff.put(row + column * columns, buffer[row + column * columns].toFloat())
			}
		}
	}

	fun store(matrix: Matrix4f) {
		check(rows == columns && rows == 4) { "This matrix has wrong dimensions: $rows by $columns" }
		store(floatBuff4)
		matrix.load(floatBuff4)
	}

	fun store(matrix: Matrix3f) {
		check(rows == columns && rows == 3) { "This matrix has wrong dimensions: $rows by $columns" }
		store(floatBuff3)
		matrix.load(floatBuff3)
	}

	fun load(buff: FloatBuffer) {
		for (row in 0 until rows) {
			for (column in 0 until columns) {
				buffer[row + column * columns] = buff[row + column * columns].toDouble()
			}
		}
	}

	fun load(matrix: Matrix4f) {
		check(rows == columns && rows == 4) { "This matrix has wrong dimensions: $rows by $columns" }
		matrix.store(floatBuff4)
		load(floatBuff4)
	}

	fun load(matrix: Matrix3f) {
		check(rows == columns && rows == 3) { "This matrix has wrong dimensions: $rows by $columns" }
		matrix.store(floatBuff3)
		load(floatBuff3)
	}

	fun resize(rows: Int, columns: Int): MutableMatrix {
		return MutableMatrix(rows, columns).also {
			for (row in 0 until this.rows.coerceAtMost(rows)) {
				for (column in 0 until this.columns.coerceAtMost(columns)) {
					it.buffer[row + column * columns] = buffer[row + column * this.columns]
				}
			}
		}
	}

	@Contract(pure = true)
	operator fun get(row: Int, column: Int): Double {
		require(row >= 0) { "Row out of bounds: $row" }
		require(row < rows) { "Row out of bounds: $row" }
		require(column >= 0) { "Column out of bounds: $row" }
		require(column < columns) { "Column out of bounds: $row" }
		return buffer[row + column * rows]
	}

	operator fun set(row: Int, column: Int, value: Double) {
		require(row >= 0) { "Row out of bounds: $row" }
		require(row < rows) { "Row out of bounds: $row" }
		require(column >= 0) { "Column out of bounds: $row" }
		require(column < columns) { "Column out of bounds: $row" }
		buffer[row + column * rows] = value
	}

	/**
	 * this = this * other
	 * при inverse = true
	 * other = this * other
	 *
	 * без создания новой матрицы
	 */
	fun multiply(other: MutableMatrix, inverse: Boolean = false): MutableMatrix {
		check(columns == other.rows) { "Matrixes are not compatible for multiplication (this: $rows by $columns, other: ${other.rows} by ${other.columns})" }
		val resultSize = rows * other.columns
		check(!inverse && buffer.size == resultSize || inverse && other.buffer.size == resultSize) { "Matrixes are not compatible for ${if (inverse) "direct" else "inverse"} multiplication in place (this: $rows by $columns, other: ${other.rows} by ${other.columns})" }

		val buff = Array(resultSize) { 0.0 }

		for (row in 0 until rows) {
			for (column in 0 until other.columns) {
				var sum = 0.0

				for (elem in 0 until columns) {
					sum += this[row, elem] * other[elem, column]
				}

				buff[row + column * columns] = sum
			}
		}

		if (inverse) {
			for (i in buff.indices) {
				other.buffer[i] = buff[i]
			}
		} else {
			for (i in buff.indices) {
				buffer[i] = buff[i]
			}
		}

		return if (inverse) other else this
	}

	/**
	 * new = this * other
	 *
	 * с созданием новой матрицы
	 */
	operator fun times(other: MutableMatrix): MutableMatrix {
		check(columns == other.rows) { "Matrixes are not compatible for multiplication (this: $rows by $columns, other: ${other.rows} by ${other.columns})" }
		val new = MutableMatrix(rows, other.columns)

		for (row in 0 until rows) {
			for (column in 0 until other.columns) {
				var sum = 0.0

				for (elem in 0 until columns) {
					sum += this[row, elem] * other[elem, column]
				}

				new[row, column] = sum
			}
		}

		return new
	}

	var translation: Vector
		get() {
			check(columns == rows && rows == 4) { "Matrix is expected to be 4x4 (this one is $rows by $columns)" }
			return Vector(this[0, 3], this[1, 3], this[2, 3])
		}

		set(value) {
			check(columns == rows && rows == 4) { "Matrix is expected to be 4x4 (this one is $rows by $columns)" }
			this[0, 0] = 1.0
			this[1, 1] = 1.0
			this[2, 2] = 1.0
			this[3, 3] = 1.0

			this[0, 3] = value.x
			this[1, 3] = value.y
			this[2, 3] = value.z
		}
}
