package ru.dbotthepony.mc.otm.menu

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Pair
import net.minecraft.core.NonNullList
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.container.iterator
import ru.dbotthepony.mc.otm.network.ExoSuitCarriedPacket
import ru.dbotthepony.mc.otm.network.ExoSuitMenuInitPacket
import ru.dbotthepony.mc.otm.network.ExoSuitSlotPacket
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel

class ExoSuitInventoryMenu(val capability: MatteryPlayerCapability) : MatteryMenu(null, CONTAINER_ID, capability.ply.inventory) {
	override val storageSlots: Collection<Slot> get() = listOf()

	val craftingGrid: CraftingContainer
	val craftingSlots: List<MatterySlot>

	init {
		if (capability.isExoSuitCraftingUpgraded) {
			craftingGrid = CraftingContainer(this, 3, 3)
		} else {
			craftingGrid = CraftingContainer(this, 2, 2)
		}

		val builder = ImmutableList.builder<MatterySlot>()

		for (i in 0 until craftingGrid.containerSize) {
			builder.add(MatterySlot(craftingGrid, i))
		}

		craftingSlots = builder.build()
	}

	val craftingResultContainer = ResultContainer()
	val craftingResultSlot = ResultSlot(capability.ply, craftingGrid, craftingResultContainer, 0, 0, 0)

	init {
		addSlot(craftingResultSlot)

		if (capability.ply is ServerPlayer) {
			addSlotListener(capability.ply.containerListener)
		}
	}

	val armorSlots: List<EquipmentSlot> = makeArmorSlots()

	val offhandSlot = object : InventorySlot(capability.ply.inventory, 40) {
		override fun getNoItemIcon(): Pair<ResourceLocation, ResourceLocation> {
			return Pair.of(InventoryMenu.BLOCK_ATLAS, InventoryMenu.EMPTY_ARMOR_SLOT_SHIELD)
		}
	}

	val allAccessibleSlots: List<Slot>

	init {
		addInventorySlots(autoFrame = false)

		craftingSlots.forEach(this::addSlot)
		addSlot(offhandSlot)
		armorSlots.forEach(this::addSlot)

		val builder = ImmutableList.builder<Slot>()

		builder.addAll(playerInventorySlots)
		builder.addAll(craftingSlots)
		builder.add(offhandSlot)

		allAccessibleSlots = builder.build()
	}

	override fun slotsChanged(container: Container) {
		super.slotsChanged(container)

		if (container == craftingGrid) {
			CraftingMenu.slotChangedCraftingGrid(this, capability.ply.level, capability.ply, craftingGrid, craftingResultContainer)
		}
	}

	override fun setSynchronizer(synchronizer: ContainerSynchronizer) {
		if (synchronizer == Companion) {
			super.setSynchronizer(synchronizer)
		}
	}

	override fun removed(player: Player) {
		if (player != capability.ply) {
			throw RuntimeException("what.")
		}

		super.removed(player)

		craftingResultContainer.clearContent()

		if (!player.level.isClientSide) {
			val iterator = craftingGrid.iterator()

			for (itemStack in iterator) {
				val leftover = moveItemStackToSlots(itemStack, playerInventorySlots)

				if (!leftover.isEmpty) {
					player.drop(leftover, true)
				}

				iterator.remove()
			}
		}
	}

	init {
		if (capability.ply is ServerPlayer) {
			setSynchronizer(Companion)
		}
	}

	override fun stillValid(player: Player): Boolean {
		return true
	}

	override fun quickMoveStack(ply: Player, slotIndex: Int): ItemStack {
		if (slots[slotIndex].item.isEmpty || !slots[slotIndex].mayPickup(ply)) {
			return ItemStack.EMPTY
		}

		if (slotIndex == craftingResultSlot.index) {
			val item = craftingResultSlot.item
			val leftover = moveItemStackToSlots(item, playerInventorySlots, simulate = true)

			if (leftover.isEmpty) {
				val copy = item.copy()
				moveItemStackToSlots(item, playerInventorySlots, simulate = false)
				item.count = 0
				craftingResultSlot.onTake(ply, copy)
				return copy
			}

			return ItemStack.EMPTY
		} else if (playerCombinedInventorySlots.any { it.index == slotIndex }) {
			val copy = slots[slotIndex].item.copy()

			if (moveItemStackTo(slots[slotIndex], playerHotbarSlots))
				return copy
			else
				return ItemStack.EMPTY
		} else if (playerHotbarSlots.any { it.index == slotIndex }) {
			val copy = slots[slotIndex].item.copy()

			if (moveItemStackTo(slots[slotIndex], playerCombinedInventorySlots))
				return copy
			else
				return ItemStack.EMPTY
		} else if (craftingSlots.any { it.index == slotIndex } || offhandSlot.index == slotIndex) {
			val copy = slots[slotIndex].item.copy()

			if (moveItemStackTo(slots[slotIndex], playerInventorySlots))
				return copy
			else
				return ItemStack.EMPTY
		}

		return ItemStack.EMPTY
	}

	override fun canTakeItemForPickAll(p_38908_: ItemStack, p_38909_: Slot): Boolean {
		return p_38909_.container != craftingResultContainer && super.canTakeItemForPickAll(p_38908_, p_38909_)
	}

	companion object : ContainerSynchronizer {
		// ServerPlayer#nextContainerCounter
		// this.containerCounter = this.containerCounter % 100 + 1;
		// значение ID никогда не будет больше 100
		const val CONTAINER_ID: Int = 101

		override fun sendInitialData(container: AbstractContainerMenu, itemStacks: NonNullList<ItemStack>, carried: ItemStack, remoteDataSlots: IntArray) {
			return sendInitialData(container as ExoSuitInventoryMenu, itemStacks, carried, remoteDataSlots)
		}

		override fun sendSlotChange(container: AbstractContainerMenu, slotId: Int, itemStack: ItemStack) {
			return sendSlotChange(container as ExoSuitInventoryMenu, slotId, itemStack)
		}

		override fun sendCarriedChange(container: AbstractContainerMenu, itemStack: ItemStack) {
			return sendCarriedChange(container as ExoSuitInventoryMenu, itemStack)
		}

		override fun sendDataChange(container: AbstractContainerMenu, dataSlotId: Int, shortData: Int) {
			return sendDataChange(container as ExoSuitInventoryMenu, dataSlotId, shortData)
		}

		fun sendInitialData(container: ExoSuitInventoryMenu, itemStacks: NonNullList<ItemStack>, carried: ItemStack, remoteDataSlots: IntArray) {
			MatteryPlayerNetworkChannel.send(container.ply as ServerPlayer, ExoSuitMenuInitPacket(itemStacks, carried, container.incrementStateId()))
		}

		fun sendSlotChange(container: ExoSuitInventoryMenu, slotId: Int, itemStack: ItemStack) {
			MatteryPlayerNetworkChannel.send(container.ply as ServerPlayer, ExoSuitSlotPacket(slotId, itemStack, container.stateId))
		}

		fun sendCarriedChange(container: ExoSuitInventoryMenu, itemStack: ItemStack) {
			MatteryPlayerNetworkChannel.send(container.ply as ServerPlayer, ExoSuitCarriedPacket(itemStack, container.stateId))
		}

		fun sendDataChange(container: ExoSuitInventoryMenu, dataSlotId: Int, shortData: Int) {
			throw UnsupportedOperationException("This should never be called. Dumbass")
		}
	}
}
